# -*- coding: utf-8 -*-
##############################################################################
#
#    Copyright (C) 2017-TODAY Eynes (<http://www.e-mips.com.ar>)
#
#    This is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from openerp import models, fields, api, exceptions, _


class account_move_line(models.Model):
    _inherit = 'account.move.line'

    pos_order_id = fields.Many2one('pos.order', string='PoS Order')


class account_bank_statement_line(models.Model):
    _inherit = 'account.bank.statement.line'

    card_number = fields.Char('Card Number')
    qty_points = fields.Float('Points Used', default=0)

    @api.multi
    def process_reconciliation(self, mv_line_dicts):
        for move_line in mv_line_dicts:
            move_line['pos_order_id'] = self.pos_statement_id.id
        res = super(account_bank_statement_line, self).process_reconciliation(mv_line_dicts)
        return res

    @api.model
    def create(self, vals):
        if vals.get('pos_statement_id'):
            journal = self.env['account.journal'].browse(vals.get('journal_id'))
            if journal.use_points:
                vals['qty_points'] = vals['amount'] * journal.points_rate
            vals['type'] = 'income'
        res = super(account_bank_statement_line, self).create(vals)
        return res

    @api.multi
    def get_credipaz_payment_code(self):
        """
        payments is browse records of account.bank.statement.line, analize the combinations and output correct code
        """
        internal_card = any(self.filtered(lambda x: (x.journal_id.pos_payment_type == 'financial') and (x.journal_id.credipaz_card is True)))
        external_card = any(self.filtered(lambda x: (x.journal_id.pos_payment_type == 'financial') and (x.journal_id.credipaz_card is False)))
        cash = any(self.filtered(lambda x: (x.journal_id.type == 'cash') and (not x.journal_id.use_points)))
        points = any(self.filtered(lambda x: x.journal_id.use_points is True))
        # Key internal_card, external_card, cash, points
        payments_dict = {
        #    intc, extca, cash,  point
            (True, False, False, False): 1,
            (True, False, True, False): 2,
            (True, False, False, True): 3,
            (True, False, True, True): 4,
            (False, False, False, True): 5,
            (False, False, True, False): 6,
            (False, False, True, True): 7,
            # 8, 9, 10, 11: credito ...
            (False, True, False, False): 12,
            (False, True, True, False): 13,
            (True, True, False, False): 14,
            (False, True, False, True): 15,
            (False, True, True, True): 16,
            (False, True, True, False): 17,
            (False, True, True, True): 18,
            # 19, 20, 21, 22: credito ...
        }
        key = (internal_card, external_card, cash, points)
        result = payments_dict.get(key, 23)  # 23 is Others
        return str(result)

