/******************************************************************************
 *    Financing - Financing on Odoo
 *    Copyright (C) 2017 Eynes
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Affero General Public License as
 *    published by the Free Software Foundation, either version 3 of the
 *    License, or (at your option) any later version.
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Affero General Public License for more details.
 *    You should have received a copy of the GNU Affero General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

function financing_models(instance, module) {

    var _t = instance.web._t;
    var round_pr = instance.web.round_precision;
    var round_di = instance.web.round_decimals;
    /**
     * Extend the order
     */
    var OrderSuper = module.Order;
    module.Order = module.Order.extend({

        initialize: function(attributes){
            var self = this;
            res = OrderSuper.prototype.initialize.call(this, attributes);
            res.receipt_type = 'invoice';
            res.quota_qty = 1;
            cf_partner_lst = self.pos.db.search_partner('Consumidor Final');
            if (cf_partner_lst.length >= 1) {
                res.set_client(cf_partner_lst[0]);
            }
            return res;
        },

        addProduct: function (product, options) {
            var self = this;
            options = options || {};

            var customer = self.get_client();
            if( !customer ){
                self.pos.pos_widget.screen_selector.show_popup('error',{
                    message: _t('Customer not selected'),
                    comment: _t('Please select a client for this order. This can be done by clicking the order tab')
                });
                return;
            }
            // else
            // TODO: try to only show enabled quotas
            //financings = self.pos.db.financing_by_product_id[product.id][customer.customer_type];
            //self.pos.update_enabled_quotas(financings, customer_type);
            var attr = JSON.parse(JSON.stringify(product));
            attr.pos = this.pos;
            attr.order = this;
            var line = new instance.point_of_sale.Orderline({}, {pos: this.pos, order: this, product: product});
            if (product.track_incoming || product.track_all) {
            // if (product.lot_unique_ok) {
                if (options.prodlot_id){  // When is a refund, and is lot traceability enabled we should have this
                    var self = this;
                    line.set_serial_id(options.prodlot_id[0]);
                    line.set_serial(options.prodlot_id[1]);
                    sr_no = options.prodlot_id[1];
                    (self.get('orderLines')).each(_.bind( function(item) {
                        if (item.get_product().lot_unique_ok && item.get_product().id == attr.id && item.get_serial() == options.prodlot_id[1]) {
                            alert('Same product is already assigned with same serial number !');
                            sr_no = null;
                            return false;
                        }
                    }, this));
                    if (sr_no != null) {
                        if(options.quantity !== undefined){
                            line.set_quantity(options.quantity);
                        }
                        if(options.price !== undefined){
                            line.set_unit_price(options.price);
                        }
                        if(options.discount !== undefined){
                            line.set_discount(options.discount);
                        }
                        var last_orderline = self.getLastOrderline();
                        if (last_orderline && last_orderline.can_be_merged_with(line) && options.merge !== false){
                            last_orderline.merge(line);
                        } else {
                            self.get('orderLines').add(line);
                        }
                        self.selectLine(self.getLastOrderline());
                    }
                } else {  // Display lot selection wizard
                    var self = this;
                    var sli = self.pos.config.stock_location_id[0];  // stock location
                    var slti = self.pos.config.stock_location_transit_id[0];  // stock transit location
                    var ctx = new instance.web.CompoundContext({'pos': true, 'product_id': attr.id, 'loc_ids': [sli, slti]})
                    new instance.web.Model("stock.production.lot").get_func("search_read")
                                (['&','&',['product_id', '=', attr.id],['quant_ids.location_id.usage','=', 'internal'],['quant_ids.location_id.id', 'in', [sli, slti]]], false, false, false, false, ctx).pipe(
                        function(result) {
                            if (result) {
                                initial_ids = _.map(result, function(x) {return x['id']});
                                this.pop = new instance.web.form.SelectCreatePopup(this);
                                    this.pop.select_element(
                                        'stock.production.lot', 
                                        {
                                            title: 'Select serial number for: ' + attr.display_name, 
                                            initial_ids: initial_ids, 
                                            initial_view: 'search',
                                            disable_multiple_selection: true,
                                            no_create: true,
                                        }, [["id", "in", initial_ids]], new instance.web.CompoundContext({}));
                            }
                            this.pop.on("elements_selected", self, function(element_ids) {
                                var dataset = new instance.web.DataSetStatic(self, 'stock.production.lot', {});
                                dataset.name_get(element_ids).done(function(data) {
                                    if (data) {
                                            new instance.web.Model("stock.production.lot").get_func("check_stock_lot")(data[0][0], ctx).pipe(
                                            function(lot_res){
                                                if (lot_res > 0) {
                                                line.set_serial_id(data[0][0]);
                                                    line.set_serial(data[0][1]);
                                                    sr_no = data[0][1];
                                                    (self.get('orderLines')).each(_.bind( function(item) {
                                                        if (item.get_product().lot_unique_ok && item.get_product().id == attr.id && item.get_serial() == data[0][1]) {
                                                            alert('Same product is already assigned with same serial number !');
                                                            sr_no = null;
                                                            return false;
                                                        }
                                                    }, this));
                                                    if (sr_no != null) {
                                                        if(options.quantity !== undefined){
                                                            line.set_quantity(options.quantity);
                                                        }
                                                        if(options.price !== undefined){
                                                            line.set_unit_price(options.price);
                                                        }
                                                        if(options.discount !== undefined){
                                                            line.set_discount(options.discount);
                                                        }
                                                        var last_orderline = self.getLastOrderline();
                                                        if (last_orderline && last_orderline.can_be_merged_with(line) && options.merge !== false){
                                                            last_orderline.merge(line);
                                                        } else {
                                                            self.get('orderLines').add(line);
                                                        }
                                                        self.selectLine(self.getLastOrderline());
                                                    }
                                                } else {
                                                    alert (_t('Not enough quantity in this serial number !'))
                                                }
                                                });
                                    }
                                });
                            });
                        });
                }
            } else {
                return OrderSuper.prototype.addProduct.call(this, product, options);
            }
        },

        addPaymentline: function(cashregister) {
            var self = this;
            var res = OrderSuper.prototype.addPaymentline.call(this, cashregister);
            //var currentOrder = self.get('selectedOrder');
            if (self.is_return_order === true) {
                return res;
            }
            self.pos.pos_widget.payment_screen.change_quota_qty(self.quota_qty);
            return res;
        },

        getCashRegisterCredipazIds: function() {  // returns ids of cashregisters with cp card
            var self = this;
            var cashregisters = self.pos.cashregisters;
            var lst = [];
            for (var i = 0, len = cashregisters.length; i < len; i++) {
                if (cashregisters[i].journal.credipaz_card) {
                    lst.push(cashregisters[i].id);
                }
            }
            return lst;
        },
        
        getCashRegisterPointsIds: function() {  // returns ids of cashregisters with points enabled
            var self = this;
            var cashregisters = self.pos.cashregisters;
            var lst = [];
            for (var i = 0, len = cashregisters.length; i < len; i++) {
                if (cashregisters[i].journal.use_points) {
                    lst.push(cashregisters[i].id);
                }
            }
            return lst;
        },

        resolve_partner_data: function(partner_data){
            var self = this;
            var partner = self.pos.db.get_partner_by_id(partner_data['partner_id']);
            partner['ws_ok'] = partner_data['ok'];
            partner['ws_points'] = partner_data['available_points'];
            partner['ws_credit'] = partner_data['available_credit'];
            partner['customer_type'] = partner_data['customer_type'];
        },

        update_partner_ui_tooltip: function (order) {
            client = order.get_client()
            cashregister_cpz_ids = order.getCashRegisterCredipazIds();
            cashregister_pts_ids = order.getCashRegisterPointsIds();
            if (client.ws_ok == false) {  // Disable cashregisters with journal.credipaz_card true
                cashregister_ids = cashregister_cpz_ids.concat(cashregister_pts_ids);
                for (var i = 0, len = cashregister_ids.length; i < len; i++) {
                    cr_id = cashregister_ids[i];
                    var selector = '.paypad-button[cash-register-id="'+cr_id+'"]'
                    cr_ui = posmodel.pos_widget.$(selector);
                    $(cr_ui).attr('disabled', 'disabled');
                }
            }
            if (client.ws_ok == true) {
                for (var i = 0, len = cashregister_pts_ids.length; i < len; i++) {
                    cr_id = cashregister_pts_ids[i];
                    var selector = '.paypad-button[cash-register-id="'+cr_id+'"]'
                    cr_ui = posmodel.pos_widget.$(selector);
                    if (parseFloat(client.ws_points) > 0) {
                        $(cr_ui).removeAttr('disabled');
                    } else {
                        $(cr_ui).attr('disabled', 'disabled');
                    }
                }
                for (var i = 0, len = cashregister_cpz_ids.length; i < len; i++) {
                    cr_id = cashregister_cpz_ids[i];
                    var selector = '.paypad-button[cash-register-id="'+cr_id+'"]'
                    cr_ui = posmodel.pos_widget.$(selector);
                    if (parseFloat(client.ws_credit) > 0) {
                        $(cr_ui).removeAttr('disabled');
                    } else {
                        $(cr_ui).attr('disabled', 'disabled');
                    }
                }
            }
            var tooltip_msg = '<div style="text-align:left;"> Puntos: ' + client.ws_points + '<br />Credito: ' + client.ws_credit + '</div>';
            var order_lst = posmodel.pos_widget.$('.orders .select-order');
            for (var i = 0, len = order_lst.length; i < len; i++) {
                order_ui = order_lst[i];
                order_seq = parseInt($(order_ui).text().trim().match(/([0-9]*)/)[0]);
                if (order_seq == order.sequence_number) {
                    $(order_ui).attr('data-original-title', tooltip_msg);
                    $(order_ui).attr('data-toggle', 'tooltip');
                    $(order_ui).tooltip({delay: {show: 50, hide: 100}});
                } 
            }
        },
        set_client: function(client){
            var self = this;
            openerp.web.blockUI();
            new instance.web.Model('res.partner').call('get_webservice_data',[client.id]).then(function(partner_data){
                self.resolve_partner_data(partner_data);
                openerp.web.unblockUI();
                return self.set('client', client);
                // self.pos.pos_widget.screen_selector.show_popup('error',{
                //     'message':_t('Warning: Cannot sell to this partner'),
                //     'comment':_t('Credipaz WS does not allow selling to this customer.'),
                // });

            },function(err,event){
                event.preventDefault();
                openerp.web.unblockUI();
                var msg = ''
                if (err.data) {
                    msg = err.data.message;
                } else {
                    msg = _t('Unknown problem while getting customer data');
                }
                self.pos.pos_widget.screen_selector.show_popup('error',{
                    'message':_t('Error: Fetching data from Credipaz WS'),
                    'comment': msg || 'Cannot get data from Credipaz WS, please check your internet connection.',
                });
            });
            this.pos.financing_engine.show_product_financing(client);
        },

        get_quotas_total: function() {
            var quotas_total = $.map(this.get('orderLines').models, 
                function(line) { return line.quota_amount }).reduce(
                    function(previousValue, currentValue){ return currentValue + previousValue; }, 0
                ); 
            //console.log(quotas_total);
            return quotas_total;
        },

        export_as_JSON: function() {
            res = OrderSuper.prototype.export_as_JSON.call(this);
            res.quota_qty = this.quota_qty;
            res.quota_amount = this.get_quotas_total();
            return res;
        }

    });

    /**
     * Extend the order line
     */
    var OrderlineSuper = module.Orderline;
    module.Orderline = module.Orderline.extend({
        initialize: function(attr, options){
            var self = this;
            res = OrderlineSuper.prototype.initialize.call(this, attr, options);
            self.quota_amount = self.get_product().list_price;
            return res;
        },

        //set_unit_price: function(price){
        //    //this.quota_amount = ;
        //    return OrderlineSuper.prototype.initialize.call(this, price);
        //},

        //get_quota_summary: function () {
        //    return this.quota_summary;
        //},

        recalculate_unit_price: function (cash_amount, financing) {
            var order = this.order;
            // var tmp_unit_price = financing.total;
            var old_prices = this.get_all_prices();
            // tmp_unit_price = round_pr(tmp_unit_price, this.pos.currency.rounding);
            var list_price = this.get_product().list_price;
            var amount_to_discount = list_price * (this.discount / 100);
            list_price -= amount_to_discount;
            // this.price = tmp_unit_price;
            // recompute all prices with using the unit price from the financing
            // var prices = this.get_all_prices();
            // var cash_amount_rate = cash_amount / prices.priceWithTax;
            // cash_amount_rate = round_di(cash_amount_rate, this.pos.dp['Financing Config.']);
            // var financed_amount_rate = 1 - cash_amount_rate;
            // financed_amount_rate = round_di(financed_amount_rate, this.pos.dp['Financing Config.']);
            // var unit_price = (tmp_unit_price * financed_amount_rate) + (list_price * cash_amount_rate);
            // unit_price = round_pr(unit_price, this.pos.currency.rounding);
            // return unit_price;
            // return financing.total;
            // NEW CODE
            // get vat amount 0.21 or 0.105
            tax_id = this.get_product().taxes_id[0];
            tax_amount = 1 + (this.pos.taxes_by_id[tax_id].amount);  // 1.21 or 1.105
            var financing_total = financing.total - (financing.total * (this.discount / 100));
            financing_rate = (financing_total / list_price) - 1;
            due_to_finance = ((old_prices['priceWithTax'] - cash_amount) / tax_amount);
            new_price = list_price + (due_to_finance * (financing_rate));
            return new_price;
        },

    });

    /**
     * Extend the POS model
     */
    var PosModelParent = module.PosModel;
    module.PosModel = module.PosModel.extend({
        /**
         * @param session
         * @param attributes
         */
        initialize: function (session, attributes) {
            var self = this;
            // fectch res.partner's customer_type
            var partner_model = _.find(this.models, function(model){
                return model.model === 'res.partner';
            });
            partner_model.fields.push('customer_type', 'ref');

            // fectch product.product's financing_tables
            var product_model = _.find(this.models, function(model){
                return model.model === 'product.product';
            });
            // product_model.fields.push('financing_tables');
            product_model.fields.push('no_interest');
            product_model.loaded = function(self, products) { self.db.add_products_with_pos(products, self) }

            // fectch account.journal's use_points, credipaz_card and pos_payment_type fields...
            // XXX: ... and we had to add name because it was not being triggered after we added the 
            // new fields (why?). It seems that when fields is an empty array it reads name field 
            // by default.
            var journal_model = _.find(this.models, function(model){
                return model.model === 'account.journal';
            });
            journal_model.fields.push('name', 'use_points', 'credipaz_card', 'points_rate', 'pos_payment_type');

            PosModelParent.prototype.initialize.apply(this, arguments);
            this.financing_engine = new module.FinancingEngine({
                'pos': this,
                'db': this.db,
                'pos_widget': this.pos_widget
            });
            get_financings(self);
            get_fp(self);
            this.enabled_quotas = [1, 3, 6, 9, 12, 18, 20];
        },

        get_financing_quotas: function (financings) {
            // get the allowed quotas and remove duplicated
            financing_quotas = financings.map(function (f) { 
              return f.quota_qty;
            }).filter(function(elem, index, self) {
                return index == self.indexOf(elem);
            });
            if ( !financing_quotas.includes(1) ) { 
                financing_quotas.push(1);
            }
            financing_quotas.sort(function(a, b) { return a - b; });
            return financing_quotas;
        },

        set_enabled_quotas: function(financings) {
            this.enabled_quotas = this.get_financing_quotas(financings);
        },

        reduce_financing_quotas: function (quotas) {
            var quotas_qty_map = {};
            quotas.map(function (qty) { 
                if (!quotas_qty_map.hasOwnProperty(qty)) { 
                   quotas_qty_map[qty] = 0;
                }
                quotas_qty_map[qty] = quotas_qty_map[qty] + 1;
            });
            return quotas_qty_map
        },

        update_enabled_quotas: function (financings, customer_type) {
            var financing_quotas = this.get_financing_quotas(financings);
            if ( financing_quotas &&
                (this.enabled_quotas.length == 0 || (this.enabled_quotas == 1 && this.enabled_quotas[0] == 1))) {
                var enabled_quotas = financing_quotas;
            }
            else {
                var currentOrder = this.get('selectedOrder');
                var orderlines = currentOrder.get('orderLines').models;
                var products = [];
                var financings = [];
                for (var i=0; i < orderlines.length; i++) {
                    var line = orderlines[i];
                    products.push(line.product);
                    var _financings = this.db.financing_by_product_id[line.product.id][customer_type];
                    financing_quotas.concat(this.get_financing_quotas(_financings))
                }
                var enabled_quotas = this.reduce_financing_quotas(financing_quotas);
            }

            this.enabled_quotas = enabled_quotas;
            this.pos_widget.payment_screen.renderElement();
        },

        /**
         * find model based on name
         * @param model_name
         * @returns {{}}
         */
        find_model: function (model_name) {
            var self = this;
            var lookup = {};
            for (var i = 0, len = self.models.length; i < len; i++) {
                if (self.models[i].model === model_name) {
                    lookup[i] = self.models[i]
                }
            }
            return lookup
        },

        //update_prices: function() {
        //    var lines = this.order.get('orderLines').models;
        //    var quota_qty = this.order.get('quota_qty');
        //    for (var i=0; i < lines.length; i++) {
        //        var line = lines[i];
        //        var financing = this.db.search_financing_by_quota_qty(
        //            line.product.id, this.customer_type, quota_qty
        //        );
        //        price = financing ? financing.quota_amount * quota_qty : 0.0;
        //        line.set_unit_price(price);
        //    };
        //},

        /**
         * @param removed_order
         * @param index
         * @param reason
         */
        on_removed_order: function (removed_order, index, reason) {
            PosModelParent.prototype.on_removed_order.apply(this, arguments);
            if ((reason === 'abandon' || removed_order.temporary) && this.get('orders').size() > 0) {
                var current_order = (this.get('orders').at(index) || this.get('orders').last());
                var partner = current_order.get_client() ?  current_order.get_client() : false;
                this.financing_engine.show_product_financing(partner);
            }
        }
    });

    // extend to add boolean with cash condition
    var PaymentlineParent = module.Paymentline;
    module.Paymentline = module.Paymentline.extend({
       initialize: function(attributes, options) {
           this.points = 0;
           var res = PaymentlineParent.prototype.initialize.apply(this, arguments);
           return res;
       },
        export_as_JSON: function(){
            var res = PaymentlineParent.prototype.export_as_JSON.apply(this);
            res['card_number'] = this.get_card_number();
            return res;
        },
        set_card_number: function(number){
            console.log(number, this);
            this.card_number = number;
        },
        get_card_number: function(){
            return this.card_number || "";
        },
        set_amount_and_points: function(value, repaint){
            this.amount = round_di(parseFloat(value) || 0, this.pos.currency.decimals);
            this.points = round_di(parseFloat(value * this.cashregister.journal.points_rate) || 0, this.pos.currency.decimals);
            var client_points = this.pos.get_order().attributes.client.ws_points || 0;
            if (this.points > client_points) {
                return this.pos.pos_widget.screen_selector.show_popup('error',{
                    'message':_t('Error: Not enough points'),
                    'comment': _t('Trying to use more points that customer has available.'),
                });
            }
            this.trigger('change:amount',this);
            repaint && this.trigger('change:selected', this);
        },
        set_points_and_amount: function(value, repaint){
            this.points = round_di(parseFloat(value) || 0, this.pos.currency.decimals);
            var client_points = this.pos.get_order().attributes.client.ws_points || 0;
            if (this.points > client_points) {
                return this.pos.pos_widget.screen_selector.show_popup('error',{
                    'message':_t('Error: Not enough points'),
                    'comment': _t('Trying to use more points that customer has available.'),
                });
            }
            this.amount = round_di(parseFloat(value / this.cashregister.journal.points_rate) || 0, this.pos.currency.decimals);
            this.trigger('change:amount',this);
            repaint && this.trigger('change:selected', this);
        },
        get_points: function(){
            var res = this.amount * this.cashregister.journal.points_rate;
            return res;
        },
        get_points_str: function(){
            return openerp.instances[this.pos.session.name].web.format_value(this.points, {
                type: 'float', digits: [69, this.pos.currency.decimals]
            });
        },
        
        //set_amount: function(value){
        //    //if ( this.cashregister.journal.pos_payment_type === 'cash' ) {
        //    //    var currentOrder = this.pos.get('selectedOrder');
        //    //    var orderlines = currentOrder.get('orderLines').models;
        //    //    if ( orderlines.length == 1 ) {
        //    //        this.pos.pos_widget.payment_screen.change_quota_qty(self.quota_qty);
        //    //    }
        //    //}
        //    return PaymentlineParent.prototype.set_amount.apply(this, arguments)
        //},
    })

    /**
     * Financing Engine to show the fiancings
     */
    module.FinancingEngine = instance.web.Class.extend({
        /**
         * @param options
         */
        init: function (options) {
            options = options || {};
            this.pos = options.pos;
            this.db = options.db;
            this.pos_widget = options.pos_widget;
        },

        set_product_ui_tooltip: function (product_ui, table) {
            var table_header = '<table class="tooltip_table">'
              + '<tr><th class="tooltip_table_header">Cuotas</th>' 
              + '<th class="tooltip_table_header">$</th>' 
              + '<th class="tooltip_table_header">Total</th>' 
              + '<th class="tooltip_table_header">T.E.A.</th>' 
              + '<th class="tooltip_table_header">C.F.T.</th></tr>';
            var closing_tag = '</table>';
            $(product_ui).find('.product-img').attr(
                'data-original-title', table_header + table + closing_tag
            );
            $(product_ui).find('.product-img').attr('data-toggle', 'tooltip');
            $(product_ui).find('.product-img').tooltip({delay: {show: 50, hide: 100}});
        },

        /**
         * @param partner
         */
        show_product_financing: function (partner) {
            if (!this.pos_widget.product_screen) return;
            var product_list_ui = this.pos_widget.product_screen.$('.product-list span.product');
            customer_type = partner ? partner.customer_type : false;
            // build financings table that will be showed
            for (var i = 0, len = product_list_ui.length; i < len; i++) {
                var product_ui = product_list_ui[i];
                var product_id = $(product_ui).data('product-id');
                var financings_table = this.db.table_by_product_id[product_id][customer_type];
                $(product_ui).find('.product-img').attr('data-original-title', financings_table);
                $(product_ui).find('.product-img').attr('data-toggle', 'tooltip');
                $(product_ui).find('.product-img').tooltip({delay: {show: 50, hide: 10}});
            };
        }
    });

    function get_fp(pos_model) {

        var res_product_pricelist = pos_model.find_model('product.pricelist');
        if (_.size(res_product_pricelist) == 1) {
            var pricelist_index = parseInt(Object.keys(res_product_pricelist)[0]);
            pos_model.models.splice(++pricelist_index, 0, 
                {
                    model: 'account.fiscal.position',
                    fields: ['name', 'id'],
                    domain: null,
                    loaded: function (self, fiscal_positions) {
                        self.fiscal_positions = fiscal_positions;
                    }
                });
        }

        // var res_partner_model = pos_model.find_model('res.partner');
        // if (_.size(res_partner_model) == 1) {
        //     var res_partner_index = parseInt(Object.keys(res_partner_model)[0]);
        //     pos_model.models[res_partner_index].fields.push('property_account_position');
        // }
    };

    function get_financings(pos_model) {
        var res_product_pricelist = pos_model.find_model('product.product');
        if (_.size(res_product_pricelist) == 1) {
            var pricelist_index = parseInt(Object.keys(res_product_pricelist)[0]);
            pos_model.models.splice(++pricelist_index, 0, 
                {
                    model: 'product.financing',
                    fields: ['product_id',
                        'customer_type',
                        'quota_qty',
                        'quota_amount',
                        'total',
                        'apr',
                        'tfc'],
                    domain: function(pos_model, tmp){ return [['pos_config_id', '=', pos_model.config.id]]; },
                    loaded: function (self, financings) {
                        self.db.add_financings(financings);
                        self.db.set_financing_tables(financings);
                    }
                });
            pos_model.models.splice(++pricelist_index, 0, {
                    model: 'financing.config',
                    fields: ['customer_type', 'quota_qty', 'monthly_rate', 'total_rate'],
                    domain: null,
                    loaded: function (self, financing_configs) {
                        self.db.add_financing_configs(financing_configs, pos_model);
                    }
                });
            };
      }
}
